@@ -1,10 +1,7 @@
-import base64
 from time import time
 from operator import itemgetter
 
 from zope.interface import implements
-from zope.app.schema.vocabulary import IVocabularyFactory
-from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
 from zope import schema
 from zope.formlib import form
 from zope.i18n import translate
@@ -23,39 +20,13 @@ from Products.PythonScripts.standard import url_quote
 from qi.portlet.TagClouds import TagCloudPortletMessageFactory as _
 
 def _cachekey(method,self):
-    #one hour cache
+    """Time, language, settings and member based cache
+    """
+    membership = getToolByName(self,'portal_membership')
     lang = self.request.get('LANGUAGE', 'en')
-    return hash((lang, time() // self.data.refreshInterval))
-
-class SubjectsVocabulary(object):
-    """Vocabulary factory for subjects.
-    This confuses me. I need this to work with non-english tags which seem to confuse zope vocabularies...
-    Need to find a more elegant way of dealing with the issue ;)
-    """
-    implements(IVocabularyFactory)
-    
-    def __call__(self, context):
-        catalog = getToolByName(context, 'portal_catalog')
-        subjects = list(catalog.uniqueValuesFor('Subject'))
-        subjects.sort();
-        terms = [SimpleTerm(value=k, token=base64.b64encode(k), title=k) for k in subjects]
-        return SimpleVocabulary(terms)
-
-SubjectsVocabularyFactory = SubjectsVocabulary()
-
-class StatesVocabulary(object):
-    """Vocabulary factory for states in the default workflow
-    """
-    implements(IVocabularyFactory)
-    
-    def __call__(self,context):
-        wt = getToolByName(context,'portal_workflow')
-        id = wt.getDefaultChain()[0]
-        defWf = wt.getWorkflowById(id)
-        terms = [SimpleTerm(value=s.id,token=s.id,title=s.title) for s in defWf.states.objectValues()]
-        return SimpleVocabulary(terms)
-        
-StatesVocabularyFactory = StatesVocabulary()
+    memberid = membership.getAuthenticatedMember()
+    return hash((lang,memberid, self.data, 
+                 time() // self.data.refreshInterval))
 
 class ITagCloudPortlet(IPortletDataProvider):
     """
@@ -78,42 +49,47 @@ class ITagCloudPortlet(IPortletDataProvider):
 
     count = schema.Int(
         title = _(u"Maximum number of shown tags."),
-        description = _(u"If greater than zero this number will limit the tags shown."),
+        description = _(u"If greater than zero this number will limit the " \
+        "tags shown."),
         required = True,
         min = 0,
         default = 0
         )
 
-    shouldRestrictBySubject = schema.Bool(
-            title = _(u"label_portletTagCloud_restrict_subjects", default=u"Restrict by keywords"),
-            description = _(u"desc_portletTagCloud_restrict_subjects",
-                            default = u"If not checked all keywords will be searched"),
-            default = False)
-
     restrictSubjects = schema.List(
         required = False,
         title = _(u"Restrict by keywords"),
-        description = _(u"Restrict the cloud keywords by searching through items filed under these keywords"),
-        value_type = schema.Choice(vocabulary='qi.portlet.TagClouds.subjects'),
+        description = _(u"Restrict the keywords searched. Leaving " \
+        "this empty will include all keywords"),
+        value_type = schema.Choice(vocabulary =
+            'qi.portlet.TagClouds.subjects'),
+        )
+    
+    filterSubjects = schema.List(
+        required = False,
+        title = _(u"Filter by keywords"),
+        description = _(u"Filter the keywords searched. Only items " \
+        "categorized with at least all the keywords selected here " \
+        "will be searched. The keywords selected here will be " \
+        "omitted from the tag clouds. Leaving the field empty will " \
+        "disable filtering"),
+        value_type = schema.Choice(vocabulary =
+            'qi.portlet.TagClouds.subjects'),
         )
-
-    shouldRestrictByType = schema.Bool(
-            title = _(u"label_portletTagCloud_restrict_objects", default=u"Resctict by type"),
-            description = _(u"desc_portletTagCloud_restrict_objects",
-                            default = u"If not checked all portal types will be searched"),
-            default = False)
 
     restrictTypes = schema.List(
         required = False,
         title = _(u"Restrict by types"),
-        description = _(u"Restrict the cloud keywords by searching through items of these types"),
-        value_type = schema.Choice(vocabulary='plone.app.vocabularies.ReallyUserFriendlyTypes'),
+        description = _(u"Restrict the content types. Leaving this empty " \
+        "will include all user-friendly content types."),
+        value_type = schema.Choice(vocabulary =
+            'plone.app.vocabularies.ReallyUserFriendlyTypes'),
         )
 
     root = schema.Choice(
             title=_(u"Root node"),
-            description=_(u"You may search for and choose a folder "
-                          "to act as the root of the navigation tree. "
+            description=_(u"You may search for and choose a folder " \
+                          "to act as the root of the navigation tree. " \
                           "Leave blank to use the Plone site root."),
             required=False,
             source=SearchableTextSourceBinder({'is_folderish' : True},
@@ -121,12 +97,15 @@ class ITagCloudPortlet(IPortletDataProvider):
     wfStates = schema.List(
             required = True,
             title = _(u"Workflow states to show"),
-            description = _(u"Which workflow states to include in the search."),
-            value_type = schema.Choice(vocabulary='qi.portlet.TagClouds.states'),)
+            description = _(u"Which workflow states to include in the " \
+                            "search."),
+            value_type = schema.Choice(vocabulary =
+                                       'qi.portlet.TagClouds.states'),)
 
     refreshInterval = schema.Int(
         title = _(u"Refresh interval"),
-        description = _(u"The maximum time in seconds for which the portal will cache the results. Be careful not to use low values."),
+        description = _(u"The maximum time in seconds for which the portal"\
+            " will cache the results. Be careful not to use low values."),
         required = True,
         min = 1,
         default = 3600,
@@ -138,13 +117,16 @@ class Assignment(base.Assignment):
 
     implements(ITagCloudPortlet)
     
-    def __init__(self, portletTitle="TagCloud", levels=5, count=0,shouldRestrictBySubject=True, restrictSubjects=[], shouldRestrictByType=True, restrictTypes=[],root=u"",wfStates=[],refreshInterval=3600):
+    def __init__(self, portletTitle="TagCloud", levels=5,
+        count=0, restrictSubjects=[], filterSubjects=[],
+        restrictTypes=[],root=u"",wfStates=[],
+        refreshInterval=3600):
+
         self.portletTitle = portletTitle
         self.levels = levels
         self.count = count
-        self.shouldRestrictBySubject = shouldRestrictBySubject
         self.restrictSubjects = restrictSubjects
-        self.shouldRestrictByType = shouldRestrictByType
+        self.filterSubjects = filterSubjects
         self.restrictTypes = restrictTypes
         self.wfStates = wfStates
         self.refreshInterval = refreshInterval
@@ -171,6 +153,8 @@ class Renderer(base.Renderer):
         self.levels = data.levels
         self.wfStates = data.wfStates
         self.count = data.count
+        self.restrictTypes = data.restrictTypes
+        self.root = data.root
 
     @ram.cache(_cachekey)
     def getTags(self):
@@ -178,7 +162,9 @@ class Renderer(base.Renderer):
         # If count has been set sort by occurences and keep the "count" first
         
         if self.count:
-            sortedOccs = sorted(tagOccs.items(), key=itemgetter(1),reverse=True)[:self.count]
+            sortedOccs = sorted(tagOccs.items(), 
+                                key=itemgetter(1),
+                                reverse=True)[:self.count]
             tagOccs = dict(sortedOccs)
             
         thresholds = self.getThresholds(tagOccs.values())
@@ -192,7 +178,20 @@ class Renderer(base.Renderer):
                 continue
             d["text"] = tag
             d["class"] = "cloud" + str(size)
-            d["href"] = self.portal_url + "/search?Subject%3Alist=" + url_quote(tag)
+            href= self.portal_url + \
+                "/search?Subject%3Alist="+url_quote(tag)
+            #Add type restrictions to search link
+            href = href+ "".join(["&portal_type%3Alist="+url_quote(ptype) 
+                for ptype in self.restrictTypes])
+            #Add workflow restrictions to search link
+            href = href+ "".join(["&review_state%3Alist="+url_quote(wstate) 
+                for wstate in self.wfStates])
+            #Add path to search link
+            if self.root:
+                href = href+"&path=%s"%getNavigationRoot(self.context,
+                    relativeRoot=self.root)
+            d["href"]=href                
+            d["count"] = translate(_(u'${count} items', mapping={'count': tagOccs[tag]}), context=self.request)
             res.append(d)
         return res
     
@@ -200,14 +199,17 @@ class Renderer(base.Renderer):
         return self.data.portletTitle
     
     def getSearchSubjects(self):
-        if (self.data.shouldRestrictBySubject) :
+        if self.data.restrictSubjects :
             result = self.data.restrictSubjects
         else:
-            result = self.catalog.uniqueValuesFor('Subject')
+            result = list(self.catalog.uniqueValuesFor('Subject'))
+        for filtertag in self.data.filterSubjects:
+            if filtertag in result:
+                result.remove(filtertag)
         return result
     
     def getSearchTypes(self):
-        if (self.data.shouldRestrictByType) :
+        if self.data.restrictTypes :
             return self.data.restrictTypes
         else :
             return self.putils.getUserFriendlyTypes()
@@ -215,15 +217,32 @@ class Renderer(base.Renderer):
     def getTagOccurrences(self):
         types = self.getSearchTypes()
         tags = self.getSearchSubjects()
+        filterTags = self.data.filterSubjects
+        
         tagOccs = {}
-        for tag in tags :
-            result = self.catalog.searchResults(Subject=tag, 
-                portal_type=types, 
-                review_state=self.wfStates,
-                path=getNavigationRoot(self.context, relativeRoot=self.data.root))          
-            length = len(result)
-            if length > 0 :
-                tagOccs[tag] = length
+        for tag in tags:
+            result = []
+            if filterTags:
+                filterquery = {'query':filterTags+[tag],
+                               'operator':'and'}
+                result = self.catalog.searchResults(
+                    Subject=filterquery, 
+                    portal_type=types, 
+                    review_state=self.wfStates,
+                    path=getNavigationRoot(self.context,
+                        relativeRoot=self.data.root)
+                    )                          
+            else:
+                result = self.catalog.searchResults(
+                    Subject=tag, 
+                    portal_type=types, 
+                    review_state=self.wfStates,
+                    path=getNavigationRoot(self.context,
+                        relativeRoot=self.data.root)
+                    )
+                           
+            if result:
+                tagOccs[tag] = len(result)
         return tagOccs
             
     def getTagSize(self, tagWeight, thresholds):
@@ -236,12 +255,15 @@ class Renderer(base.Renderer):
         return size
     
     def getThresholds(self, sizes):
+        """This algorithm was taken from Anders Pearson's blog:
+         http://thraxil.com/users/anders/posts/2005/12/13/scaling-tag-clouds/
+        """
         if not sizes:
             return [1 for i in range(0, self.levels)]
         minimum =  min(sizes)
         maximum = max(sizes)
-        #this algorithm was taken from Anders Pearson's blog: http://thraxil.com/users/anders/posts/2005/12/13/scaling-tag-clouds/
-        return [pow(maximum - minimum + 1, float(i) / float(self.levels)) for i in range(0, self.levels)]
+        return [pow(maximum - minimum + 1, float(i) / float(self.levels)) 
+            for i in range(0, self.levels)]
     
     @property
     def available(self):
